'************************************************************************************************************
'                      DRIVER FOR MULTIPLE TEMPERATURE SENSORS DS18B20 ON THE SAME 1-WIRE BUS

'Driver name: "multi_ds18b20.inc"

'Command: Ds18b20_identify        For DS18B20 sensor identification with CRC for reading control ID

'Command: Ds18b20                 Start temperature conversion of all used sensors, read Scratchpad with CRC
                                ' reading control and calculate sensors temperature
'Command: Ds18b20_id x            x = 1 to Nr_of_ds18b20, read ID and save into internal EEPROM,
                                ' used only by Ds18b20_identify command
'Notes:

'Uses LCD to display data.

'After the End command in the user program include the CRC data table "crc8.inc"

'Mladen Cempuh, Date:08.2024.
'------------------------------------------------------------------------------------------------------------

Const Max_ds18b20 = 8                                       ' max. possible number of supported DS18B20 sensors

Const Bytes_for_id = 64                                     ' Bytes_for_id = Max_ds18b20 * 8

#if Nr_of_ds18b20 < 2
           "For only one DS18B20 sensor uses ds18b20.inc"   ' mark compilation error if  Nr_of_ds18b20 < 2
#endif

#if Nr_of_ds18b20 > Max_ds18b20
               "Too many DS18B20 sensors"                   ' mark compilation error if  Nr_of_ds18b20 > Max_ds18b20
#endif

'RAM variables:
Dim Ds18b20_nr As Byte                                      ' Byte variable for the ordinal number of the sensor
Dim Crc_id As Byte                                          ' Byte variable for ID CRC , 0 --> ID CRC OK ; 1 -->ID CRC NOT OK
Dim Crc(nr_of_ds18b20) As Byte                              ' for CRC of used sensors , =0 --> CRC OK ; >0 --> CRC NOT OK
Dim Ds18b20$id(8) As Byte                                   ' 8 bytes array for read ROM data of the selected sensor
Dim Ds$sc(9) As Byte                                        ' 9 bytes array for read Scratchpad bytes for selected DS18B20
Dim Temperature_ds(nr_of_ds18b20) As Single                 ' one Single variablein array for temperature of every used DS18B20
Dim Probe_nr As Byte                                        ' Byte variable for the ordinal number of the sensor during identification

'RAM auxiliary variables:
Dim I$ds As Byte
Dim Tmp$ds As Byte
Dim T$ds As Integer
Dim Id$start As Byte
Dim Id$stop As Byte

'Internal EEPROM variables:
Dim Not_used As Eram Byte                                   ' not use because address 0 in internal EEPROM is not reliable
Dim Id_eeprom(bytes_for_id) As Eram Byte                    'Byte array variables in internal EEPROM for save sensor ROM data of all possible sensors

Config Submode = New                                        ' no need to Declare Sub commands

Goto End_multi_ds18b20                                      ' exit from "multi_ds18b20.inc"

'=================== Command: Ds18b20_id x ==================

' x --> ordinal number of sensor DS18B20

'Read DS18B20 ID  and save ID into internal EEPROM
'Connect only one DS18B20 sensor for ID identification
'That command is used only by the Ds18b20 command

Sub Ds18b20_id(byval Probe As Byte)

Ds18b20_nr = Probe

If Ds18b20_nr > Nr_of_ds18b20 Then Goto End_ds18b20_id

1wreset                                                     ' reset the 1-Wire bus
1wwrite &H33                                                ' read rom
If Err = 1 Then Goto End_ds18b20_id                         ' error --> not find device DS18B20 on the 1-Wire bus, exit command Ds18b20_id x

For I$ds = 1 To 8                                           ' read sensor ROM into 8 bytes RAM array Ds18b20$id
Ds18b20$id(i$ds) = 1wread()
Next

If Ds18b20$id(8) = Crc8(ds18b20$id(1) , 7) Then             ' calculated CRC and control is received CRC match

 Crc_id = 0                                                 ' Crc_id = 0 --> crc for read ID (sensor ROM data) is correct

Else

 Crc_id = 1                                                 '  Crc_id = 1 --> crc for read ID (sensor ROM data) is not correct
 Goto End_ds18b20_id                                        ' then exit from command Ds18b20_id x

End If

Gosub Id$_start_stop                                        ' go to subrutine Id$_start_stop for calculated Id$start and Id$stop place in internal
                                                             ' EEPROM byte array Ds18b20$id_eeprom depending of ordinal number of sensor
Tmp$ds = 1
                                                             'Id$start is starting place in EEPROM array Id_eeprom for write first byte of sensor ROM
For I$ds = Id$start To Id$stop                              'Id$stop is ending place in EEPROM array Id_eeprom for write last byte of sensor ROM
 Id_eeprom(i$ds) = Ds18b20$id(tmp$ds)                       ' write 8 bytes of sensor ROM from RAM array Ds18b20$id into EEPROM array Id_eeprom
 Incr Tmp$ds
Next

Goto End_ds18b20_id                                         ' exit command Ds18b20_id x

Id$_start_stop:                                             ' subrutine for calculated start and stop position in EEPROM array Ds18b20$id_eeprom
Id$stop = Ds18b20_nr * 8                                    ' for saving ID in EEPROM depending of Byte variable Ds18b20_nr
Id$start = Id$stop - 7
Return

End_ds18b20_id:

End Sub

'==================== Command: Ds18b20 ======================

'Start temperature convert for all used DS18B20 sensors
'and actual measuring and calculate temperature of all used
'DS18B20 sensors with CRC control

Sub Ds18b20

'Convert temperature for all DS18B20
1wreset                                                     ' reset the bus
1wwrite &HCC                                                ' skip ROM
1wwrite &H44                                                ' start convert for all DS18B20 on the 1-Wire bus
Waitms 800                                                  ' wait 800ms for a reliable end of conversion

'Match and read all used DS18b20 sensors

Ds18b20_nr = 1                                              ' ordinal number for DS18B20 sensor is 1

Lab$read_ds18b20:

Gosub Id$_start_stop                                        ' go to subrutine  Id$_start_stop for calculated Id$start and Id$stop position in
                                                             'EEPROM array Id_eeprom depending of ordinal number of sensor
Tmp$ds = 1

For I$ds = Id$start To Id$stop
 Ds18b20$id(tmp$ds) = Id_eeprom(i$ds)                       ' read 8 bytes ID for selected sensor from EEPROM array Id_eeprom into RAM
 Incr Tmp$ds                                                ' 8 bytes array Ds18b20$id
Next

1wreset                                                     ' reset the bus
1wwrite &H55                                                ' match rom

For I$ds = 1 To 8                                           ' from the RAM array Ds18b20$id send on the 1-Wire bus 8 bytes ID
 1wwrite Ds18b20$id(i$ds)                                   ' for selected DS18B20 sensor (selected by ordinal number)                          '
Next

If Err = 1 Then Goto Lab$_end_ds18b20                       ' Err = 1 --> no sensor was found on the 1-Wire bus with that ID then exit command Ds18b20

1wwrite &HBE                                                ' read 9 data bytes from Scratchpad of selected DS18B20 sensor
Ds$sc(1) = 1wread(9)                                        ' in array Ds$sc

 Crc_calc                                                   ' calculation CRC for Scratchpad of selected DS18B20 sensor

If Crc(ds18b20_nr) = 0 Then                                 ' if it is OK, calculate Temperature_ds with 0.0625C resolution
 T$ds = Makeint(ds$sc(1) , Ds$sc(2))
 Temperature_ds(ds18b20_nr) = T$ds / 16
End If

If Ds18b20_nr < Nr_of_ds18b20 Then
 Incr Ds18b20_nr
 Goto Lab$read_ds18b20
End If

Lab$_end_ds18b20:

End Sub

'-------------------------------------------------------------
Sub Crc_calc                                                ' calculate 8 bit CRC

'Faster CRC8 calculation using 256 byte table from "crc8.inc"
Crc(ds18b20_nr) = 0
For I$ds = 1 To 9
 Tmp$ds = Crc(ds18b20_nr) Xor Ds$sc(i$ds)
 Crc(ds18b20_nr) = Lookup(tmp$ds , Crc8)
Next

End Sub

'================ Command: Ds18b20_identify =================

'Select the sensor number, read the ID and saves it
'in the internal EEPROM.Uses LCD to display data.

'Remark: Connect only one DS18B20 sensor at the time.
'------------------------------------------------------------

Sub Ds18b20_identify

Probe_nr = 1                                                ' starting ordinal number for this sensor is 1

Cls : Lcd "*IDENTIFICATION*"
Locate 2 , 1 : Lcd "Only one DS18B20" : Wait 5

Cls : Lcd "DS18B20 probe ?"

Select_probe_nr:

Locate 2 , 1 : Lcd "Nr. = " ; Probe_nr

Debounce Up$ , 0 , Probe_nr_incr , Sub                      ' with Up key go to subrutine Ds18b20_nr_incr  for increment ordinal number

If Enter$ = 0 Then Goto Ds18b20_id_read_save                ' with Enter key go to label Ds18b20_id_read_save for read and save ID (sensor ROM data)
                                                              ' into internal EEPROM
Goto Select_probe_nr

Probe_nr_incr:                                              ' every press on UP key increment ordinal number of sensor (roll from 1 to Nr_of ds18b20)
 Incr Probe_nr
 If Probe_nr > Nr_of_ds18b20 Then Probe_nr = 1
Return

Ds18b20_id_read_save:

 Ds18b20_id Probe_nr                                        ' read ROM data from DS18B20 sensor and save into internal EEPROM for the selected sensor

Cls

 If Err = 1 Then                                            ' no DS18B20 sensor was found on the 1-Wire bus
  Lcd "DS18B20 Nr." ; Probe_nr
  Locate 2 , 1 : Lcd " NOT FOUND!"
  Goto Lab_ Wait_reset
 End If

 If Crc_id = 0 Then
  Lcd "DS18B20 Nr." ; Probe_nr ; " ID:"
  Locate 2 , 1                                              ' if CRC is correct show complete ID with CRC on LCD
  For I$ds = 1 To 8
  Lcd Hex(ds18b20$id(i$ds))
  Next
 Else
  Lcd "ID CRC NOT OK!"                                      ' write message "ID CRC NOT OK!" on LCD if CRC is not correct
End If

Lab_wait_reset:

 Goto Lab_wait_reset                                        ' wait for microcontroller reset

End Sub

'***************************************************************

End_multi_ds18b20: